package net.chipkillmar.poker;

import java.util.*;

/**
 * Contains the logic for determining the winner between 2 5-card poker hands, and for ranking a poker hand.
 *
 * @author Chip Killmar
 */
public class Logic {
    private static final CardComparator cardComparator = new CardComparator();

    /**
     * Determines the winner between two 5-card poker hands.
     *
     * @param hand1 The first poker hand.
     * @param hand2 The second poker hand.
     * @return 1 if hand1 wins; -1 if hand2 wins; 0 if there is a tie.
     */
    public static int determineWinner(Card[] hand1, Card[] hand2) {
        Tuple2<HandRanking, List<Value>> rank1 = rank(hand1);
        Tuple2<HandRanking, List<Value>> rank2 = rank(hand2);
        if (rank1.getFirst() != rank2.getFirst()) {
            return rank1.getFirst().ordinal() > rank2.getFirst().ordinal() ? 1 : -1;
        } else {
            for (int i = 0; i < rank1.getSecond().size(); ++i) {
                int ordinal1 = rank1.getSecond().get(i).ordinal();
                int ordinal2 = rank2.getSecond().get(i).ordinal();
                if (ordinal1 > ordinal2) {
                    return 1;
                } else if (ordinal1 < ordinal2) {
                    return -1;
                }
            }
            return 0;
        }
    }

    /**
     * Ranks a 5-card poker hand.
     *
     * @param hand A poker hand.
     * @return A 2-tuple comprising the hand ranking, e.g. "Four of a Kind", and a compressed list containing the
     *         relevant values for the hand ranking in descending order, e.g. "Ace, King" for a four-of-a-kind with a
     *         King kicker.
     */
    public static Tuple2<HandRanking, List<Value>> rank(Card[] hand) {
        checkValid(hand);

        Arrays.sort(hand, cardComparator);

        List<Value> flush = isFlush(hand);
        Value straight = isStraight(hand);
        if (flush != null && straight != null) {
            return new Tuple2<HandRanking, List<Value>>(HandRanking.STRAIGHT_FLUSH, Arrays.asList(straight));
        }

        List<Value> fourOfAKind = isFourOfAKind(hand);
        if (fourOfAKind != null) {
            return new Tuple2<HandRanking, List<Value>>(HandRanking.FOUR_OF_A_KIND, fourOfAKind);
        }

        List<Value> fullHouse = isFullHouse(hand);
        if (fullHouse != null) {
            return new Tuple2<HandRanking, List<Value>>(HandRanking.FULL_HOUSE, fullHouse);
        }

        if (flush != null) {
            return new Tuple2<HandRanking, List<Value>>(HandRanking.FLUSH, flush);
        }

        if (straight != null) {
            return new Tuple2<HandRanking, List<Value>>(HandRanking.STRAIGHT, Arrays.asList(straight));
        }

        List<Value> threeOfAKind = isThreeOfAKind(hand);
        if (threeOfAKind != null) {
            return new Tuple2<HandRanking, List<Value>>(HandRanking.THREE_OF_A_KIND, threeOfAKind);
        }

        List<Value> twoPair = isTwoPair(hand);
        if (twoPair != null) {
            return new Tuple2<HandRanking, List<Value>>(HandRanking.TWO_PAIR, twoPair);
        }

        List<Value> onePair = isOnePair(hand);
        if (onePair != null) {
            return new Tuple2<HandRanking, List<Value>>(HandRanking.ONE_PAIR, onePair);
        }

        return new Tuple2<HandRanking, List<Value>>(HandRanking.HIGH_CARD, Arrays.asList(hand[4].getValue(), hand[3].getValue(), hand[2].getValue(), hand[1].getValue(), hand[0].getValue()));
    }

    private static List<Value> isFlush(Card[] hand) {
        Suit suit = hand[4].getSuit();
        List<Value> values = new ArrayList<Value>();
        values.add(hand[4].getValue());
        for (int i = 3; i >= 0; --i) {
            if (hand[i].getSuit() != suit) {
                return null;
            } else {
                values.add(hand[i].getValue());
            }
        }
        return values;
    }

    private static Value isStraight(Card[] hand) {
        if (hand[0].getValue() == Value.TWO && hand[1].getValue() == Value.THREE && hand[2].getValue() == Value.FOUR && hand[3].getValue() == Value.FIVE && hand[4].getValue() == Value.ACE) {
            return hand[3].getValue();
        }
        Value value = hand[0].getValue();
        for (int i = 1; i < hand.length; ++i) {
            Value tempValue = hand[i].getValue();
            if (tempValue.ordinal() != value.ordinal() + 1) {
                return null;
            } else {
                value = tempValue;
            }
        }
        return hand[4].getValue();
    }

    private static List<Value> isFullHouse(Card[] hand) {
        if (hand[0].getValue() == hand[1].getValue() && hand[0].getValue() == hand[2].getValue()) {
            if (hand[3].getValue() == hand[4].getValue()) {
                return Arrays.asList(hand[0].getValue(), hand[3].getValue());
            }
        }

        if (hand[2].getValue() == hand[3].getValue() && hand[2].getValue() == hand[4].getValue()) {
            if (hand[0].getValue() == hand[1].getValue()) {
                return Arrays.asList(hand[2].getValue(), hand[0].getValue());
            }
        }

        return null;
    }

    private static List<Value> isTwoPair(Card[] hand) {
        if (hand[0].getValue() == hand[1].getValue()) {
            if (hand[2].getValue() == hand[3].getValue()) {
                return Arrays.asList(hand[2].getValue(), hand[0].getValue(), hand[4].getValue());
            }

            if (hand[3].getValue() == hand[4].getValue()) {
                return Arrays.asList(hand[3].getValue(), hand[0].getValue(), hand[2].getValue());
            }
        }

        if (hand[1].getValue() == hand[2].getValue()) {
            if (hand[3].getValue() == hand[4].getValue()) {
                return Arrays.asList(hand[3].getValue(), hand[1].getValue(), hand[0].getValue());
            }
        }

        return null;
    }

    private static List<Value> isFourOfAKind(Card[] hand) {
        if (hand[0].getValue() == hand[1].getValue() && hand[1].getValue() == hand[2].getValue() && hand[2].getValue() == hand[3].getValue()) {
            return Arrays.asList(hand[0].getValue(), hand[4].getValue());
        }
        if (hand[1].getValue() == hand[2].getValue() && hand[2].getValue() == hand[3].getValue() && hand[3].getValue() == hand[4].getValue()) {
            return Arrays.asList(hand[1].getValue(), hand[0].getValue());
        }
        return null;
    }

    private static List<Value> isThreeOfAKind(Card[] hand) {
        if (hand[0].getValue() == hand[1].getValue() && hand[1].getValue() == hand[2].getValue()) {
            return Arrays.asList(hand[0].getValue(), hand[4].getValue(), hand[3].getValue());
        }
        if (hand[1].getValue() == hand[2].getValue() && hand[2].getValue() == hand[3].getValue()) {
            return Arrays.asList(hand[1].getValue(), hand[4].getValue(), hand[0].getValue());
        }
        if (hand[2].getValue() == hand[3].getValue() && hand[3].getValue() == hand[4].getValue()) {
            return Arrays.asList(hand[2].getValue(), hand[1].getValue(), hand[0].getValue());
        }
        return null;
    }

    private static List<Value> isOnePair(Card[] hand) {
        if (hand[0].getValue() == hand[1].getValue()) {
            return Arrays.asList(hand[0].getValue(), hand[4].getValue(), hand[3].getValue(), hand[2].getValue());
        }
        if (hand[1].getValue() == hand[2].getValue()) {
            return Arrays.asList(hand[1].getValue(), hand[4].getValue(), hand[3].getValue(), hand[0].getValue());
        }
        if (hand[2].getValue() == hand[3].getValue()) {
            return Arrays.asList(hand[2].getValue(), hand[4].getValue(), hand[1].getValue(), hand[0].getValue());
        }
        if (hand[3].getValue() == hand[4].getValue()) {
            return Arrays.asList(hand[3].getValue(), hand[2].getValue(), hand[1].getValue(), hand[0].getValue());
        }
        return null;
    }

    private static void checkValid(Card[] hand) {
        if (hand == null) {
            throw new NullPointerException("hand");
        }
        if (hand.length != 5) {
            throw new IllegalArgumentException("A hand must have 5 cards.");
        }
        for (Card card : hand) {
            if (card == null) {
                throw new IllegalArgumentException("A hand must have 5 cards.");
            }
        }
        Set<Card> cardSet = new HashSet<Card>(Arrays.asList(hand));
        if (cardSet.size() != 5) {
            throw new IllegalArgumentException("A hand must have 5 unique cards.");
        }
    }

    private Logic() {
    }
}